import pytest

from aces_high_core import Card, Rank, Suit


def test_card_has_suit():
    card = Card(Suit.CLUBS, Rank.TWO)
    assert card.suit == Suit.CLUBS


def test_card_has_rank():
    card = Card(Suit.CLUBS, Rank.THREE)
    assert card.rank == Rank.THREE


def test_card_has_repr():
    card = Card(Suit.HEARTS, Rank.KING)
    assert repr(card) == "The King of Hearts"


@pytest.mark.parametrize(
    "rank,expected_value",
    [
        (Rank.ACE, 1),
        (Rank.TWO, 2),
        (Rank.THREE, 3),
        (Rank.FOUR, 4),
        (Rank.FIVE, 5),
        (Rank.SIX, 6),
        (Rank.SEVEN, 7),
        (Rank.EIGHT, 8),
        (Rank.NINE, 9),
        (Rank.TEN, 10),
        (Rank.JACK, 11),
        (Rank.QUEEN, 12),
        (Rank.KING, 13),
    ],
)
def test_card_has_correct_value(rank, expected_value):
    card = Card(Suit.CLUBS, rank)
    assert card.value == expected_value



import random
from abc import ABC, abstractmethod

from .card import Card, Rank, Suit


class BaseDeck(ABC):
    def __init__(self):
        self._cards = []
        self._initialize_cards()

    def __len__(self):
        return len(self._cards)

    def reset(self):
        self._initialize_cards()

    @abstractmethod
    def shuffle(self):
        pass

    @abstractmethod
    def deal(self, n=1):
        pass

    @abstractmethod
    def _initialize_cards(self):
        pass


class StandardDeck(BaseDeck):
    def shuffle(self):
        random.shuffle(self._cards)

    def deal(self, n: int = 1):
        if n > len(self):
            raise ValueError("Deck does not have enough cards to deal")
        if n <= 0:
            raise ValueError("Must deal one or more cards")

        cards_to_deal = [self._cards.pop() for _ in range(n)]
        return cards_to_deal

    def _initialize_cards(self):
        for suit in Suit:
            for rank in Rank:
                self._cards.append(Card(suit, rank))



from dataclasses import dataclass
from enum import Enum


class Suit(Enum):
    CLUBS = "Clubs"
    DIAMONDS = "Diamonds"
    HEARTS = "Hearts"
    SPADES = "Spades"


class Rank(Enum):
    ACE = "Ace"
    TWO = "Two"
    THREE = "Three"
    FOUR = "Four"
    FIVE = "Five"
    SIX = "Six"
    SEVEN = "Seven"
    EIGHT = "Eight"
    NINE = "Nine"
    TEN = "Ten"
    JACK = "Jack"
    QUEEN = "Queen"
    KING = "King"


face_values = {
    Rank.ACE: 1,
    Rank.TWO: 2,
    Rank.THREE: 3,
    Rank.FOUR: 4,
    Rank.FIVE: 5,
    Rank.SIX: 6,
    Rank.SEVEN: 7,
    Rank.EIGHT: 8,
    Rank.NINE: 9,
    Rank.TEN: 10,
    Rank.JACK: 11,
    Rank.QUEEN: 12,
    Rank.KING: 13,
}


@dataclass(frozen=True, eq=True)
class Card:
    suit: Suit
    rank: Rank

    @property
    def value(self):
        return face_values[self.rank]

    def __repr__(self):
        return f"The {self.rank.value} of {self.suit.value}"



import pytest

from aces_high_core import Card, Rank, StandardDeck, Suit


@pytest.fixture
def standard_deck():
    return StandardDeck()


def test_deck_starts_with_52_cards(standard_deck):
    assert len(standard_deck) == 52


def test_deck_contains_all_clubs(standard_deck):
    for rank in Rank:
        assert Card(Suit.CLUBS, rank) in standard_deck._cards


def test_deck_contains_all_hearts(standard_deck):
    for rank in Rank:
        assert Card(Suit.HEARTS, rank) in standard_deck._cards


def test_deck_contains_all_spades(standard_deck):
    for rank in Rank:
        assert Card(Suit.SPADES, rank) in standard_deck._cards


def test_deck_contains_all_diamonds(standard_deck):
    for rank in Rank:
        assert Card(Suit.DIAMONDS, rank) in standard_deck._cards


def test_deck_shuffles_itself(standard_deck):
    original_deck = [card for card in standard_deck._cards]
    standard_deck.shuffle()
    assert original_deck != standard_deck._cards


# this test ensures that even if some cards are in the same position post-shuffle, not all of them are
def test_shuffle_moves_most_cards(standard_deck):
    original_deck = standard_deck._cards.copy()
    standard_deck.shuffle()
    # Count how many cards remain in the same position
    unchanged = sum(1 for o, s in zip(original_deck, standard_deck._cards) if o == s)
    assert unchanged < len(original_deck)


# this test ensures a shuffled deck still contains all of the same cards
def test_shuffle_preserves_all_cards(standard_deck):
    original_set = set(standard_deck._cards)
    standard_deck.shuffle()
    assert set(standard_deck._cards) == original_set


def test_deal_returns_one_card(standard_deck):
    dealt = standard_deck.deal()
    assert len(dealt) == 1
    assert len(standard_deck) == 51


def test_deal_returns_expected_number_of_cards(standard_deck):
    dealt = standard_deck.deal(5)
    assert len(dealt) == 5
    assert len(standard_deck) == 47


def test_cannot_deal_more_cards_than_deck_has(standard_deck):
    with pytest.raises(ValueError):
        standard_deck.deal(53)


def test_cannot_deal_zero_cards(standard_deck):
    with pytest.raises(ValueError):
        standard_deck.deal(0)


def test_cannot_deal_negative_cards(standard_deck):
    with pytest.raises(ValueError):
        standard_deck.deal(-1)


def test_cannot_deal_from_empty_deck(standard_deck):
    standard_deck.deal(52)
    with pytest.raises(ValueError):
        standard_deck.deal(1)
